@@ -317,7 +317,7 @@ class UnconstrainedMonotonicTransform(MonotonicTransform):
     r"""Creates a monotonic transformation :math:`f(x)` by integrating a positive
     univariate function :math:`g(x)`.
 
-    .. math:: f(x) = \int_0^x g(t) ~ dt + C
+    .. math:: f(x) = \int_0^x g(u) ~ du + C
 
     The definite integral is estimated by a :math:`n`-point Gauss-Legendre quadrature.
 
@@ -337,7 +337,7 @@ class UnconstrainedMonotonicTransform(MonotonicTransform):
         self,
         g: Callable[[Tensor], Tensor],
         C: Tensor,
-        n: int = 32,
+        n: int = 16,
         **kwargs,
     ):
         super().__init__(self.f, **kwargs)
@@ -358,6 +358,44 @@ class UnconstrainedMonotonicTransform(MonotonicTransform):
         return self.g(x).log()
 
 
+class SOSPolynomialTransform(UnconstrainedMonotonicTransform):
+    r"""Creates a sum-of-squares (SOS) polynomial transformation.
+
+    The transformation :math:`f(x)` is expressed as the primitive integral of the
+    sum of :math:`K` squared polynomials of degree :math:`L`.
+
+    .. math:: f(x) = \int_0^x \sum_{i = 1}^K
+        \left( 1 + \sum_{j = 0}^L a_{i,j} ~ u^j \right)^2 ~ du + C
+
+    References:
+        | Sum-of-Squares Polynomial Flow (Jaini et al., 2019)
+        | https://arxiv.org/abs/1905.02325
+
+    Arguments:
+        a: The polynomial coefficients :math:`a`, with shape :math:`(*, K, L + 1)`.
+        C: The integration constant :math:`C`.
+        kwargs: Keyword arguments passed to :class:`UnconstrainedMonotonicTransform`.
+    """
+
+    domain = constraints.real
+    codomain = constraints.real
+    bijective = True
+    sign = +1
+
+    def __init__(self, a: Tensor, C: Tensor, **kwargs):
+        super().__init__(self.g, C, a.shape[-1], **kwargs)
+
+        self.a = a
+        self.i = torch.arange(a.shape[-1]).to(a.device)
+
+    def g(self, x: Tensor) -> Tensor:
+        x = x / self.bound
+        x = x[..., None] ** self.i
+        p = 1 + self.a @ x[..., None]
+
+        return p.squeeze(dim=-1).square().sum(dim=-1)
+
+
 class AutoregressiveTransform(Transform):
     r"""Transform via an autoregressive mapping.
 
@@ -381,8 +419,19 @@ class AutoregressiveTransform(Transform):
         self.meta = meta
         self.passes = passes
 
+        self._cache = None, None
+
     def _call(self, x: Tensor) -> Tensor:
-        return self.meta(x)(x)
+        _x, _f = self._cache
+
+        if x is _x:
+            f = _f
+        else:
+            f = self.meta(x)
+
+        self._cache = x, f
+
+        return f(x)
 
     def _inverse(self, y: Tensor) -> Tensor:
         x = torch.zeros_like(y)
@@ -392,7 +441,16 @@ class AutoregressiveTransform(Transform):
         return x
 
     def log_abs_det_jacobian(self, x: Tensor, y: Tensor) -> Tensor:
-        return self.meta(x).log_abs_det_jacobian(x, y).sum(dim=-1)
+        _x, _f = self._cache
+
+        if x is _x:
+            f = _f
+        else:
+            f = self.meta(x)
+
+        self._cache = x, f
+
+        return f.log_abs_det_jacobian(x, y).sum(dim=-1)
 
 
 class PermutationTransform(Transform):
